@@ -1,15 +1,14 @@
 ;;; tmr.el --- Set timers using a convenient notation -*- lexical-binding: t -*-
 
-;; Copyright (C) 2020-2022  Free Software Foundation, Inc.
+;; Copyright (C) 2020-2024  Free Software Foundation, Inc.
 
 ;; Author: Protesilaos Stavrou <info@protesilaos.com>,
 ;;         Damien Cassou <damien@cassou.me>,
 ;;         Daniel Mendler <mail@daniel-mendler.de>
-;; Maintainer: TMR Development <~protesilaos/tmr@lists.sr.ht>
-;; URL: https://git.sr.ht/~protesilaos/tmr
-;; Mailing-List: https://lists.sr.ht/~protesilaos/tmr
-;; Version: 0.4.0
-;; Package-Requires: ((emacs "27.1"))
+;; Maintainer: Protesilaos Stavrou <info@protesilaos.com>
+;; URL: https://github.com/protesilaos/tmr
+;; Version: 1.0.0
+;; Package-Requires: ((emacs "27.1") (compat "30.0.0.0"))
 ;; Keywords: convenience, timer
 
 ;; This file is NOT part of GNU Emacs.
@@ -37,11 +36,15 @@
 
 ;;; Code:
 
+(require 'compat)
 (require 'seq)
 (eval-when-compile (require 'cl-lib))
 
 (defgroup tmr ()
   "TMR May Ring: set timers using a simple notation."
+  :link '(info-link :tag "Info Manual" "(tmr)")
+  :link '(url-link :tag "Homepage" "https://protesilaos.com/emacs/tmr")
+  :link '(emacs-library-link :tag "Library Source" "tmr.el")
   :group 'data)
 
 (define-obsolete-variable-alias
@@ -56,9 +59,20 @@ variable that holds a list of strings.
 
 The default value of `tmr-description-history', is the name of a
 variable that contains input provided by the user at the relevant
-prompt of the `tmr' and `tmr-with-description' commands."
+prompt of the `tmr' and `tmr-with-details' commands."
   :type '(choice symbol (repeat string)))
 
+(defcustom tmr-notification-urgency 'normal
+  "The urgency level of the desktop notification.
+Values can be `low', `normal' (default), or `critical'.
+
+The desktop environment or notification daemon is responsible for
+such notifications."
+  :type '(choice
+          (const :tag "Low" low)
+          (const :tag "Normal" normal)
+          (const :tag "Critical" critical)))
+
 (defcustom tmr-sound-file
   "/usr/share/sounds/freedesktop/stereo/alarm-clock-elapsed.oga"
   "Path to sound file used by `tmr-sound-play'.
@@ -86,23 +100,24 @@ Each function must accept a timer as argument."
   :type 'hook
   :options '(tmr-print-message-for-created-timer))
 
-(declare-function tmr-notification-notify "ext:tmr-notification.el" (title message))
-
 (define-obsolete-variable-alias
   'tmr-timer-completed-functions
   'tmr-timer-finished-functions
   "0.4.0")
 
 (defcustom tmr-timer-finished-functions
-  (list #'tmr-print-message-for-finished-timer
-        #'tmr-sound-play
-        #'tmr-notification-notify)
+  (list #'tmr-sound-play
+        #'tmr-notification-notify
+        #'tmr-print-message-for-finished-timer
+        #'tmr-acknowledge-minibuffer)
   "Functions to execute when a timer is finished.
 Each function must accept a timer as argument."
   :type 'hook
-  :options (list #'tmr-print-message-for-finished-timer
-                 #'tmr-sound-play
-                 #'tmr-notification-notify))
+  :options (list #'tmr-sound-play
+                 #'tmr-notification-notify
+                 #'tmr-print-message-for-finished-timer
+                 #'tmr-acknowledge-minibuffer
+                 #'tmr-acknowledge-dialog))
 
 (defcustom tmr-timer-cancelled-functions
   (list #'tmr-print-message-for-cancelled-timer)
@@ -110,6 +125,53 @@ Each function must accept a timer as argument."
 Each function must accept a timer as argument."
   :type 'hook)
 
+(defcustom tmr-finished-indicator "✔"
+  "Indicator for a finished timer."
+  :package-version '(tmr . "1.0.0")
+  :type 'string)
+
+(defgroup tmr-faces ()
+  "Faces for `tmr'."
+  :link '(info-link :tag "Info Manual" "(tmr)")
+  :link '(url-link :tag "Homepage" "https://protesilaos.com/emacs/tmr")
+  :link '(emacs-library-link :tag "Library Source" "tmr.el")
+  :group 'tmr)
+
+(defface tmr-duration nil
+  "Face for styling the duration of a timer."
+  :package-version '(tmr . "1.0.0")
+  :group 'tmr-faces)
+
+(defface tmr-description '((t :inherit bold))
+  "Face for styling the description of a timer."
+  :package-version '(tmr . "1.0.0")
+  :group 'tmr-faces)
+
+(defface tmr-start-time '((t :inherit success))
+  "Face for styling the start time of a timer."
+  :package-version '(tmr . "1.0.0")
+  :group 'tmr-faces)
+
+(defface tmr-end-time '((t :inherit error))
+  "Face for styling the start time of a timer."
+  :package-version '(tmr . "1.0.0")
+  :group 'tmr-faces)
+
+(defface tmr-is-acknowledged '((t :inherit success))
+  "Face for styling the acknowledgment confirmation."
+  :package-version '(tmr . "1.0.0")
+  :group 'tmr-faces)
+
+(defface tmr-must-be-acknowledged '((t :inherit warning))
+  "Face for styling the acknowledgment confirmation."
+  :package-version '(tmr . "1.0.0")
+  :group 'tmr-faces)
+
+(defface tmr-finished '((t :inherit error))
+  "Face for styling the confirmation of a finished timer."
+  :package-version '(tmr . "1.0.0")
+  :group 'tmr-faces)
+
 (cl-defstruct (tmr--timer
                (:constructor tmr--timer-create)
                (:copier tmr--timer-copy))
@@ -125,6 +187,10 @@ Each function must accept a timer as argument."
    nil
    :read-only nil
    :documentation "Non-nil if the timer is finished.")
+  (acknowledgep
+   nil
+   :read-only nil
+   :documentation "Non-nil if the timer must be acknowledged.")
   (timer-object
    nil
    :read-only nil
@@ -148,15 +214,24 @@ Each function must accept a timer as argument."
     ;; enough to be used when starting a timer but also when cancelling
     ;; one: check `tmr-print-message-for-created-timer' and
     ;; `tmr-print-message-for-cancelled-timer'.
-    (format "TMR start %s; end %s; %s %s%s"
-            (propertize start 'face 'success)
-            (propertize end 'face 'error)
-            (if (string-match-p ":" (tmr--timer-input timer))
+    (format "TMR start %s; end %s; %s %s%s%s"
+            (propertize start 'face 'tmr-start-time)
+            (propertize end 'face 'tmr-end-time)
+            (if (string-search ":" (tmr--timer-input timer))
                 "until"
               "duration")
-            (tmr--timer-input timer)
+            (propertize (tmr--timer-input timer) 'face 'tmr-duration)
+            (cond
+             ((and (tmr--timer-acknowledgep timer)
+                   (tmr--timer-finishedp timer))
+              (concat "; " (propertize "acknowledged" 'face 'tmr-is-acknowledged)))
+             ((tmr--timer-acknowledgep timer)
+              (concat "; " (propertize "acknowledge" 'face 'tmr-must-be-acknowledged)))
+             ((tmr--timer-finishedp timer)
+              (concat "; " (propertize "finished" 'face 'tmr-finished)))
+             (t ""))
             (if description
-                (format " [%s]" (propertize description 'face 'bold))
+                (concat "; " (propertize description 'face 'tmr-description))
               ""))))
 
 (defun tmr--long-description-for-finished-timer (timer)
@@ -168,10 +243,10 @@ optional `tmr--timer-description'."
         (description (tmr--timer-description timer)))
     ;; For the TMR prefix, see comment in `tmr--long-description'.
     (format "TMR Time is up!\n%s%s %s\n%s %s"
-            (if description (format "%s\n" description) "")
-            (propertize "Started" 'face 'success)
+            (if description (concat (propertize description 'face 'tmr-description) "\n") "")
+            (propertize "Started" 'face 'tmr-start-time)
             start
-            (propertize "Ended" 'face 'error)
+            (propertize "Ended" 'face 'tmr-end-time)
             end)))
 
 (defun tmr--format-creation-date (timer)
@@ -185,28 +260,31 @@ optional `tmr--timer-description'."
 (defun tmr--format-remaining (timer)
   "Format remaining time of TIMER."
   (if (tmr--timer-finishedp timer)
-      "✔"
-    (let ((secs (round (- (float-time (tmr--timer-end-date timer))
-                          (float-time)))))
-      (if (> secs 3600)
-          (format "%sh %sm" (/ secs 3600) (/ (% secs 3600) 60))
-        (if (> secs 60)
-            (format "%sm %ss" (/ secs 60) (% secs 60))
-          (format "%ss" secs))))))
+      tmr-finished-indicator
+    (let* ((secs (round (- (float-time (tmr--timer-end-date timer))
+                           (float-time))))
+           (str (if (> secs 3600)
+                    (format "%sh %sm" (/ secs 3600) (/ (% secs 3600) 60))
+                  (if (> secs 60)
+                      (format "%sm %ss" (/ secs 60) (% secs 60))
+                    (format "%ss" secs)))))
+      (if (< secs 0)
+          ;; Negative remaining time occurs for non-acknowledged timers with
+          ;; additional duration.
+          (propertize str 'face 'tmr-must-be-acknowledged)
+        str))))
 
 (defun tmr--format-time (time)
   "Return a human-readable string representing TIME."
   (format-time-string "%T" time))
 
-(defun tmr--unit (now time)
-  "Determine common time unit for TIME given current time NOW."
+(defun tmr--parse-duration (now time)
+  "Parse TIME string given current time NOW."
   (save-match-data
     (cond
-     ((natnump time)
-      (* time 60))
-     ((and (stringp time) (string-match-p "\\`[0-9]+\\(?:\\.[0-9]+\\)?\\'" time))
+     ((string-match-p "\\`[0-9]+\\(?:\\.[0-9]+\\)?\\'" time)
       (* (string-to-number time) 60))
-     ((and (stringp time) (string-match "\\`\\([0-9]+\\):\\([0-9]+\\)\\(?::\\([0-9]+\\)\\)?\\'" time))
+     ((string-match "\\`\\([0-9]+\\):\\([0-9]+\\)\\(?::\\([0-9]+\\)\\)?\\'" time)
       (let ((val (decode-time now)))
         (setf (decoded-time-hour val) (string-to-number (match-string 1 time))
               (decoded-time-minute val) (string-to-number (match-string 2 time))
@@ -217,7 +295,7 @@ optional `tmr--timer-description'."
         (when (time-less-p val now)
           (user-error "Time %s is already over" time))
         (ceiling (float-time (time-subtract val now)))))
-     ((and (stringp time) (string-match "\\`\\([0-9]+\\(?:\\.[0-9]+\\)?\\)[mhs]\\'" time))
+     ((string-match "\\`\\([0-9]+\\(?:\\.[0-9]+\\)?\\)[mhs]\\'" time)
       (let ((num (string-to-number (match-string 1 time))))
         (pcase (aref time (1- (length time)))
           (?s num)
@@ -232,7 +310,6 @@ Populated by `tmr' and then operated on by `tmr-cancel'.")
 (defvar tmr--update-hook nil
   "Hooks to execute when timers are changed.")
 
-;;;###autoload
 (defun tmr-remove (timer)
   "Cancel and remove TIMER object set with `tmr' command.
 Interactively, let the user choose which timer to cancel with
@@ -243,7 +320,6 @@ completion."
   (run-hooks 'tmr--update-hook)
   (run-hook-with-args 'tmr-timer-cancelled-functions timer))
 
-;;;###autoload
 (defun tmr-cancel (timer)
   "Cancel TIMER object set with `tmr' command.
 Interactively, let the user choose which timer to cancel with
@@ -252,7 +328,6 @@ chooses only among active timers."
   (interactive (list (tmr--read-timer "Cancel timer: " :active)))
   (tmr-remove timer))
 
-;;;###autoload
 (defun tmr-reschedule (timer)
   "Reschedule TIMER.
 This is the same as cloning it, prompting for duration and
@@ -262,7 +337,6 @@ cancelling the original one."
   (let (tmr-timer-cancelled-functions)
     (tmr-cancel timer)))
 
-;;;###autoload
 (defun tmr-edit-description (timer description)
   "Change TIMER description with that of DESCRIPTION."
   (interactive
@@ -272,7 +346,14 @@ cancelling the original one."
   (setf (tmr--timer-description timer) description)
   (run-hooks 'tmr--update-hook))
 
-;;;###autoload
+(defun tmr-toggle-acknowledge (timer)
+  "Toggle ackowledge flag of TIMER."
+  (interactive
+   (list
+    (tmr--read-timer "Toggle acknowledge flag of timer: ")))
+  (setf (tmr--timer-acknowledgep timer) (not (tmr--timer-acknowledgep timer)))
+  (run-hooks 'tmr--update-hook))
+
 (defun tmr-remove-finished ()
   "Remove all finished timers."
   (interactive)
@@ -321,6 +402,25 @@ If there are no timers, throw an error."
         (or (and selected (get-text-property 0 'tmr-timer selected))
             (user-error "No timer selected")))))))
 
+(declare-function notifications-notify "notifications" (&rest params))
+
+(defun tmr-notification-notify (timer)
+  "Dispatch a notification for TIMER.
+
+Read: (info \"(elisp) Desktop Notifications\") for details."
+  (if (featurep 'dbusbind)
+      (let ((title "TMR May Ring (Emacs tmr package)")
+            (body (tmr--long-description-for-finished-timer timer)))
+        (unless (fboundp 'notifications-notify)
+          (require 'notifications))
+        (notifications-notify
+         :title title
+         :body body
+         :app-name "GNU Emacs"
+         :urgency tmr-notification-urgency
+         :sound-file tmr-sound-file))
+    (warn "Emacs has no DBUS support, TMR notifications unavailable")))
+
 ;; NOTE 2022-04-21: Emacs has a `play-sound' function but it only
 ;; supports .wav and .au formats.  Also, it does not work on all
 ;; platforms and Emacs needs to be compiled --with-sound capabilities.
@@ -359,11 +459,9 @@ TIMER is unused."
 If DEFAULT is provided, use that as a default."
   (let ((def (or default (nth 0 tmr-duration-history))))
     (read-string
-     (if def
-         (format "N minutes for timer (append `h' or `s' for other units) [%s]: " def)
-       "N minutes for timer (append `h' or `s' for other units): ")
-     nil
-     'tmr-duration-history def)))
+     (format-prompt
+      "N minutes for timer (append `h' or `s' for other units)" def)
+     nil 'tmr-duration-history def)))
 
 (defvar tmr-description-history '()
   "Minibuffer history of `tmr' descriptions.")
@@ -372,15 +470,56 @@ If DEFAULT is provided, use that as a default."
   "Helper prompt for descriptions in `tmr'.
 If optional DEFAULT is provided use it as a default candidate."
   (completing-read
-   (if default
-       (format "Description for this tmr [%s]: " default)
-     "Description for this tmr: ")
+   (format-prompt "Description for this tmr" default)
    (tmr--completion-table
     (if (listp tmr-description-list)
         tmr-description-list
       (symbol-value tmr-description-list)))
    nil nil nil 'tmr-description-history default))
 
+(defun tmr--acknowledge-prompt ()
+  "Ask the user if a timer must be acknowledged."
+  (y-or-n-p "Acknowledge timer after finish? "))
+
+(defun tmr-acknowledge-dialog (timer)
+  "Acknowledge TIMER by showing a GUI dialog."
+  (when-let (((tmr--timer-acknowledgep timer))
+             (duration
+              (x-popup-dialog
+               t
+               `(,(tmr--long-description-for-finished-timer timer)
+                 ("Acknowledge" . nil)
+                 ("+ 1 m" . 60)
+                 ("+ 5 m" . 300)
+                 ("+ 10 min" . 600)
+                 ("+ 15 min" . 960)
+                 nil))))
+    (tmr--continue-overtime timer duration)))
+
+(defun tmr-acknowledge-minibuffer (timer)
+  "Acknowledge TIMER using the minibuffer."
+  (when (tmr--timer-acknowledgep timer)
+    (while
+        (let ((input
+               (read-from-minibuffer
+                (concat (tmr--long-description-for-finished-timer timer)
+                        "\nAcknowledge with `ack' or additional duration: "))))
+          (not (or (equal input "ack")
+                   (when-let ((duration
+                               (ignore-errors
+                                 (tmr--parse-duration (current-time) input))))
+                     (tmr--continue-overtime timer duration)
+                     t)))))))
+
+(defun tmr--continue-overtime (timer duration)
+  "Continue TIMER even after it expired for DURATION.
+This function is used if a timer is not acknowledged."
+  (setf (tmr--timer-finishedp timer) nil
+        (tmr--timer-timer-object timer)
+        (run-with-timer duration nil #'tmr--complete timer))
+  (run-hooks 'tmr--update-hook)
+  (run-hook-with-args 'tmr-timer-created-functions timer))
+
 (defun tmr--complete (timer)
   "Mark TIMER as finished and execute `tmr-timer-finished-functions'."
   (setf (tmr--timer-finishedp timer) t)
@@ -388,7 +527,7 @@ If optional DEFAULT is provided use it as a default candidate."
   (run-hook-with-args 'tmr-timer-finished-functions timer))
 
 ;;;###autoload
-(defun tmr (time &optional description)
+(defun tmr (time &optional description acknowledgep)
   "Set timer to TIME duration and notify after it elapses.
 
 When TIME is a number, it is interpreted as a count of minutes.
@@ -400,51 +539,63 @@ With optional DESCRIPTION as a prefix (\\[universal-argument]),
 prompt for a description among `tmr-description-list', though
 allow for any string to serve as valid input.
 
+With optional ACKNOWLEDGEP non-nil the timer must be acknowledged
+after it finished, such that the timer cannot be missed.
+
 This command also plays back `tmr-sound-file' if it is available.
 
 To cancel the timer, use the `tmr-cancel' command.
 
 To always prompt for a DESCRIPTION when setting a timer, use the
-command `tmr-with-description' instead of this one."
+command `tmr-with-details' instead of this one."
   (interactive
    (list
     (tmr--read-duration)
-    (when current-prefix-arg (tmr--description-prompt))))
+    (when current-prefix-arg (tmr--description-prompt))
+    (when current-prefix-arg (tmr--acknowledge-prompt))))
+  (when (natnump time)
+    (setq time (number-to-string time)))
   (let* ((creation-date (current-time))
-         (duration (tmr--unit creation-date time))
+         (duration (tmr--parse-duration creation-date time))
          (timer (tmr--timer-create
                  :description description
+                 :acknowledgep acknowledgep
                  :creation-date creation-date
                  :end-date (time-add creation-date duration)
-                 :input time))
-         (timer-object (run-with-timer
-                        duration nil
-                        #'tmr--complete timer)))
-    (setf (tmr--timer-timer-object timer) timer-object)
+                 :input time)))
+    (setf (tmr--timer-timer-object timer)
+          (run-with-timer duration nil #'tmr--complete timer))
     (push timer tmr--timers)
     (run-hooks 'tmr--update-hook)
     (run-hook-with-args 'tmr-timer-created-functions timer)))
 
 ;;;###autoload
-(defun tmr-with-description (time description)
-  "Set timer to TIME duration and notify with DESCRIPTION after it elapses.
+(defun tmr-with-details (time &optional description acknowledgep)
+  "Set timer to TIME duration and notify after it elapses.
 
-See `tmr' for a description of the arguments.  The difference
-between the two commands is that `tmr-with-description' always
-asks for a description whereas `tmr' only asks for it when the
+See `tmr' for a description of the arguments DESCRIPTION and
+ACKNOWLEDGEP.  The difference between the two commands is that
+`tmr-with-details' always asks for a description and if the timer
+should be acknowledged whereas `tmr' only asks for it when the
 user uses a prefix argument (\\[universal-argument])."
   (interactive
    (list
     (tmr--read-duration)
-    (tmr--description-prompt)))
-  (tmr time description))
+    (tmr--description-prompt)
+    (tmr--acknowledge-prompt)))
+  (tmr time description acknowledgep))
+
+(define-obsolete-function-alias
+  'tmr-with-description
+  'tmr-with-details
+  "0.4.0")
 
-;;;###autoload
 (defun tmr-clone (timer &optional prompt)
   "Create a new timer by cloning TIMER.
 With optional PROMPT, such as a prefix argument, ask for
 confirmation about the duration.  When PROMPT is a double prefix
-argument, ask for a description as well.
+argument, ask for a description as well and ask if the timer must
+be acknowledged.
 
 Without a PROMPT, clone TIMER outright."
   (interactive
@@ -457,19 +608,60 @@ Without a PROMPT, clone TIMER outright."
      (format "%s" (tmr--timer-input timer)))
    (if (equal prompt '(16))
        (tmr--description-prompt (tmr--timer-description timer))
-     (tmr--timer-description timer))))
+     (tmr--timer-description timer))
+   (if (equal prompt '(16))
+       (tmr--acknowledge-prompt)
+     (tmr--timer-acknowledgep timer))))
 
 (defun tmr--completion-table (candidates &optional category annotation)
   "Make completion table for CANDIDATES with sorting disabled.
 CATEGORY is the completion category.
 ANNOTATION is an annotation function."
-   (lambda (str pred action)
-     (if (eq action 'metadata)
-         `(metadata (display-sort-function . identity)
-                    (cycle-sort-function . identity)
+  (lambda (str pred action)
+    (if (eq action 'metadata)
+        `(metadata (display-sort-function . identity)
+                   (cycle-sort-function . identity)
                    (annotation-function . ,annotation)
-                    (category . ,category))
-       (complete-with-action action candidates str pred))))
+                   (category . ,category))
+      (complete-with-action action candidates str pred))))
+
+(defvar-keymap tmr-action-map
+  :doc "Action map for TMRs, which can be utilized by Embark."
+  "k" #'tmr-remove
+  "r" #'tmr-remove
+  "R" #'tmr-remove-finished
+  "c" #'tmr-clone
+  "a" #'tmr-toggle-acknowledge
+  "e" #'tmr-edit-description
+  "s" #'tmr-reschedule)
+
+(defvar-keymap tmr-prefix-map
+  :doc "Global prefix map for TMRs.
+This map should be bound to a global prefix."
+  "+" #'tmr
+  "*" #'tmr-with-details
+  "t" #'tmr
+  "T" #'tmr-with-details
+  "l" 'tmr-tabulated-view ;; autoloaded
+  "c" #'tmr-clone
+  "s" #'tmr-reschedule
+  "a" #'tmr-toggle-acknowledge
+  "e" #'tmr-edit-description
+  "r" #'tmr-remove
+  "R" #'tmr-remove-finished
+  "k" #'tmr-cancel)
+
+;;;###autoload (autoload 'tmr-prefix-map "tmr" nil t 'keymap)
+(defalias 'tmr-prefix-map tmr-prefix-map)
+
+(defvar embark-keymap-alist)
+(defvar embark-post-action-hooks)
+(with-eval-after-load 'embark
+  (add-to-list 'embark-keymap-alist '(tmr-timer . tmr-action-map))
+  (cl-loop
+   for cmd the key-bindings of tmr-action-map
+   if (commandp cmd) do
+   (add-to-list 'embark-post-action-hooks (list cmd 'embark--restart))))
 
 (provide 'tmr)
 ;;; tmr.el ends here
